﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;

namespace HeuristicLab.BioBoost.Evaluators {

  /// <summary>
  /// Object pool for reusing frequently created and discarded objects.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  public class ObjectPool<T> where T : new() {

    private readonly Stack<T> pool;

    public int MaxPoolSize { get; set; }

    public Func<T> Create;
    public Action<T> Clear;
    public Func<T, int, bool> Accept; 

    private static void DoNothing(T o) {}
    private static bool DefaultAccept(T o, int count) { return count < 100; }

    public ObjectPool() : this(() => new T(), DoNothing, DefaultAccept) {}

    public ObjectPool(Func<T> create, Action<T> clear, Func<T, int, bool> accept) {
      pool = new Stack<T>();
      Create = create;
      Clear = clear;
      Accept = accept;
    }

    public T Get() {
      lock (this) {
        if (pool.Count > 0) {
          return pool.Pop();
        }
      }
      return Create();
    }

    public T GetCleared() {
      lock (this) {
        if (pool.Count > 0) {
          var val = pool.Pop();
          Clear(val);
          return val;
        }
      }
      return Create();
    }

    public void Return(T o) {
      lock (this) {
        if (Accept(o, pool.Count))
          pool.Push(o);
      }
    }

    public void Reset() {
      lock (this)
        pool.Clear();
    }

  }
}
